@@ -1,29 +1,39 @@
+from __future__ import unicode_literals
+
 from copy import deepcopy
 
-from .util import X, LocalParams, make_fq
+from .util import X, LocalParams, make_fq, process_value, _pop_from_kwargs
+from .types import instantiate, get_to_python, Boolean
+from .compat import force_unicode
 
 
+DEFAULT_OP_SEP = '__'
+DEFAULT_VAL_SEP = ':'
+
 def between_op(f, v):
-    v1, v2 = v.split(':')
-    return X(**{'%s__between' % f: (v1, v2)})
+    v1, v2 = v.split(DEFAULT_VAL_SEP)
+    return X(**{'{}__between'.format(f): (v1, v2)})
+
 
 def isnull_op(f, v):
     if v == '1':
-        return X({'%s__isnull' % f: True})
+        return X({'{}__isnull'.format(f): True})
     elif v == '0':
-        return X({'%s__isnull' % f: False})
+        return X({'{}__isnull'.format(f): False})
     return
 
+
 OPERATORS = {
-    'exact': lambda f, v: X(**{'%s__exact' % f: v}),
-    'gte': lambda f, v: X(**{'%s__gte' % f: v}),
-    'gt': lambda f, v: X(**{'%s__gt' % f: v}),
-    'lte': lambda f, v: X(**{'%s__lte' % f: v}),
-    'lt': lambda f, v: X(**{'%s__lt' % f: v}),
+    'exact': lambda f, v: X(**{'{}__exact'.format(f): v}),
+    'gte': lambda f, v: X(**{'{}__gte'.format(f): v}),
+    'gt': lambda f, v: X(**{'{}__gt'.format(f): v}),
+    'lte': lambda f, v: X(**{'{}__lte'.format(f): v}),
+    'lt': lambda f, v: X(**{'{}__lt'.format(f): v}),
     'between': between_op,
     'isnull': isnull_op,
     }
 
+
 class QueryFilter(object):
     def __init__(self, *filters):
         self.filters = []
@@ -64,10 +74,10 @@ class QueryFilter(object):
         return query
 
     def add_filter(self, filter):
-        if isinstance(filter, basestring):
+        if not isinstance(filter, BaseFilter):
             filter = Filter(filter)
-        if filter.name in self.filter_names:
-            return
+        # if filter.name in self.filter_names:
+        #     return
         self.filters.append(filter)
         self.filter_names.append(filter.name)
 
@@ -83,51 +93,52 @@ class QueryFilter(object):
         self.results = results
         for f in self.filters:
             f.process_results(self.results, self.params)
-            
-            # for fq, local_params in results.query._fq:
-            #     if local_params.get('tag') == f.name:
-            #         print fq
 
-            # for facet in results.facet_queries:
-            #     if facet.local_params.get('ex') == f.name:
-            #         print facet
 
 class BaseFilter(object):
 
-    def __init__(self, name, coerce=None):
+    def __init__(self, name, type=None):
         self.name = name
-        self.coerce = coerce
+        self.type = instantiate(type)
+        self.to_python = (
+            self.type.process_param_value
+            if hasattr(self.type, 'process_param_value') else
+            (lambda v: v)
+        )
 
     def _filter_and_split_params(self, params):
         """Returns [(operator1, value1), (operator2, value2)]"""
         items = []
         for p, v in sorted(params.items(), key=lambda i: i[0]):
-            ops = p.split('__')
+            ops = p.split(DEFAULT_OP_SEP)
             name = ops[0]
             if len(ops) == 1:
                 op = 'exact'
             else:
-                op = '__'.join(ops[1:])
+                op = DEFAULT_OP_SEP.join(ops[1:])
             if name == self.name:
+                if not isinstance(v, (list, tuple)):
+                    v = [v]
                 for w in v:
-                    if self.coerce:
-                        try:
-                            w = self.coerce(w)
-                        except ValueError:
-                            continue
+                    try:
+                        w = self.to_python(w)
+                    except ValueError:
+                        continue
                     yield op, w
 
     def process_results(self, results, params):
         raise NotImplementedError()
 
+
 class Filter(BaseFilter):
     fq_connector = X.OR
 
-    def __init__(self, name, field=None, coerce=None, _local_params=None,
+    def __init__(self, name, field=None, type=None, local_params=None,
                  select_multiple=True, default=None, **kwargs):
-        super(Filter, self).__init__(name, coerce=coerce)
+        super(Filter, self).__init__(name, type=type)
         self.field = field or self.name
-        self.local_params = LocalParams(_local_params)
+        self.local_params = LocalParams(
+            kwargs.pop('_local_params', local_params))
         self.select_multiple = select_multiple
         self.default = default
     
@@ -139,8 +150,8 @@ class Filter(BaseFilter):
                 xs = [x for x, lp in fqs if x]
                 if xs:
                     return query.filter(*xs,
-                                         _op=self.fq_connector,
-                                         _local_params=local_params)
+                                        _local_params=local_params,
+                                        _op=self.fq_connector)
             else:
                 x, lp = fqs[-1]
                 local_params = LocalParams(lp)
@@ -166,14 +177,8 @@ class Filter(BaseFilter):
     def process_results(self, results, params):
         pass
 
-class FacetFilterValue(object):
-    def __init__(self, filter_name, facet_value, selected, title=None, **kwargs):
-        self.filter_name = filter_name
-        self.facet_value = facet_value
-        self.selected = selected
-        self._title = title
-        self.opts = kwargs
 
+class FacetFilterValueMixin(object):
     def __unicode__(self):
         return unicode(self.title)
 
@@ -189,6 +194,10 @@ class FacetFilterValue(object):
     def value(self):
         return self.facet_value.value
 
+    @property
+    def param_value(self):
+        return process_value(self.value, safe=True)
+
     @property
     def count(self):
         return self.facet_value.count
@@ -197,17 +206,28 @@ class FacetFilterValue(object):
     def instance(self):
         return self.facet_value.instance
 
+    
+class FacetFilterValue(FacetFilterValueMixin):
+    def __init__(self, filter_name, facet_value, selected, title=None, **kwargs):
+        self.filter_name = filter_name
+        self.facet_value = facet_value
+        self.selected = selected
+        self._title = title
+        self.opts = kwargs
+
+
 class FacetFilter(Filter):
     filter_value_cls = FacetFilterValue
     
-    def __init__(self, name, field=None, filter_value_cls=None,
-                 _local_params=None, _instance_mapper=None,
+    def __init__(self, name, field=None, filter_value_cls=None, type=None,
+                 local_params=None, instance_mapper=None,
                  select_multiple=True, **kwargs):
-        super(FacetFilter, self).__init__(name, field,
+        super(FacetFilter, self).__init__(name, field, type=type,
                                           select_multiple=select_multiple)
         self.filter_value_cls = filter_value_cls or self.filter_value_cls
-        self.local_params = LocalParams(_local_params)
-        self._instance_mapper = _instance_mapper
+        self.local_params = LocalParams(
+            kwargs.pop('_local_params', local_params))
+        self._instance_mapper = kwargs.pop('_instance_mapper', instance_mapper)
         self.kwargs = kwargs
         self.values = []
         self.selected_values = []
@@ -233,11 +253,12 @@ class FacetFilter(Filter):
     def apply(self, query, params):
         query = super(FacetFilter, self).apply(query, params)
         local_params = LocalParams(self.local_params)
-        local_params['ex'] = self.name
         local_params['key'] = self.name
+        local_params['ex'] = self.name
         query = query.facet_field(
             self.field, _local_params=local_params,
-            _instance_mapper=self.instance_mapper, **self.kwargs)
+            _instance_mapper=self.instance_mapper, _type=self.type,
+            **self.kwargs)
         return query
 
     def process_results(self, results, params):
@@ -252,11 +273,124 @@ class FacetFilter(Filter):
                     self.add_value(self.filter_value_cls(self.name, fv, selected))
                 break
 
+
+class FacetPivotFilterValueMixin(object):
+    @property
+    def all_values(self):
+        return self.pivot.all_values
+
+    @property
+    def selected_values(self):
+        return self.pivot.selected_values
+
+    @property
+    def values(self):
+        return self.pivot.values
+
+
+class FacetPivotFilterValue(FacetFilterValueMixin, FacetPivotFilterValueMixin):
+    def __init__(self, filter_name, facet_values, selected, title=None,
+                 **kwargs):
+        self.filter_name = filter_name
+        self.facet_values = facet_values
+        self.facet_value = facet_values[-1]
+        self.selected = selected
+        self._title = title
+        self.opts = kwargs
+        self.pivot = None
+
+    @property
+    def param_value(self):
+        return DEFAULT_VAL_SEP.join(process_value(fv.value, safe=True) for fv in self.facet_values)
+    
+
+class FacetPivotFilter(FacetFilter):
+    filter_value_cls = FacetPivotFilterValue
+
+    def __init__(self, name, field=None, filter_value_cls=None,
+                 type=None, _pivot_filter=None, **kwargs):
+        super(FacetPivotFilter, self).__init__(
+            name, field=field, filter_value_cls=filter_value_cls,
+            type=type, **kwargs)
+        self._pivot_filter = _pivot_filter
+
+    def bind(self, pivot_filter):
+        return FacetPivotFilter(
+            self.name, field=self.field, filter_value_cls=self.filter_value_cls,
+            type=self.type, _pivot_filter=pivot_filter, **self.kwargs)
+        
+    def process_facet(self, facet, pivot_filters, selected_values, facet_values):
+        cur_selected_values = set(self.to_python(v[0]) for v in selected_values)
+
+        for fv in facet.values:
+            selected = fv.value in cur_selected_values
+            filter_value = self.filter_value_cls(
+                self._pivot_filter.name, facet_values + [fv], selected)
+            self.add_value(filter_value)
+            if fv.pivot:
+                pivot_filter = pivot_filters[0].bind(self._pivot_filter)
+                pivot_filter.process_facet(
+                    fv.pivot,
+                    pivot_filters[1:],
+                    [v[1:] for v in selected_values if len(v) > 1 and fv.value == v[0]],
+                    facet_values + [fv])
+                filter_value.pivot = pivot_filter
+                
+
+class PivotFilter(BaseFilter, FacetPivotFilterValueMixin):
+    def __init__(self, name, *pivots, **kwargs):
+        super(PivotFilter, self).__init__(name)
+        self._pivots = pivots
+        self._pivot_fields = [p.field for p in pivots]
+        self._pivot_kwargs = []
+        for p in self._pivots:
+            kw = p.kwargs.copy()
+            kw.update(
+                _instance_mapper=p._instance_mapper,
+                _type=p.type,
+            )
+            self._pivot_kwargs.append(kw)
+        self.local_params = LocalParams(_pop_from_kwargs(kwargs, 'local_params'))
+        self.pivot = self._pivots[0].bind(self)
+
+    def apply(self, query, params):
+        local_params = LocalParams(self.local_params)
+        local_params['key'] = self.name
+        local_params['ex'] = self.name
+        query = query.facet_pivot(
+            *zip(self._pivot_fields, self._pivot_kwargs),
+            _local_params=local_params)
+        fqs = []
+        for op, value in self._filter_and_split_params(params):
+            value = force_unicode(value)
+            op_func = OPERATORS.get(op)
+            if op_func:
+                x = X()
+                for field, v in zip(self._pivot_fields, value.split(DEFAULT_VAL_SEP)):
+                    x = x & op_func(field, v)
+                fqs.append(x)
+        local_params = LocalParams()
+        local_params['tag'] = self.name
+        if fqs:
+            query = query.filter(X(*fqs, _op='OR'), _local_params=local_params)
+        return query
+
+    def process_results(self, results, params):
+        selected_values = [force_unicode(v).split(DEFAULT_VAL_SEP)
+                           for v in params.get(self.name, [])]
+        self.pivot.process_facet(
+            results.get_facet_pivot(self.name),
+            self._pivots[1:],
+            selected_values,
+            [])
+
+                
 class FacetQueryFilterValue(object):
-    def __init__(self, name, fq, _local_params=None, title=None, **kwargs):
+    def __init__(self, name, fq, local_params=None, title=None, **kwargs):
         self.filter_name = None
         self.value = name
-        self.local_params = LocalParams(_local_params)
+        self.local_params = LocalParams(
+            kwargs.pop('_local_params', local_params))
         self.fq = fq
         self.title = title
         self.opts = kwargs
@@ -268,12 +402,13 @@ class FacetQueryFilterValue(object):
                     
     @property
     def _key(self):
-        return '%s__%s' % (self.filter_name, self.value)
+        return '{}__{}'.format(self.filter_name, self.value)
 
     @property
     def count(self):
         return self.facet_query.count
 
+
 class FacetQueryFilter(Filter):
     def __init__(self, name, *args, **kwargs):
         super(FacetQueryFilter, self).__init__(name, **kwargs)
@@ -327,12 +462,13 @@ class FacetQueryFilter(Filter):
         if not was_selected and self.default:
             self.get_value(self.default).selected = True
 
+
 class RangeFilter(Filter):
     fq_connector = X.AND
     
-    def __init__(self, name, field=None, coerce=int,
+    def __init__(self, name, field=None, type=None,
                  gather_stats=False, exclude_filter=True, **kwargs):
-        super(RangeFilter, self).__init__(name, field, coerce=coerce, **kwargs)
+        super(RangeFilter, self).__init__(name, field, type=type, **kwargs)
         self.gather_stats = gather_stats
         self.exclude_filter = exclude_filter
         self.from_value = None
@@ -375,12 +511,12 @@ class OrderingValue(object):
     DESC = 'desc'
     
     def __init__(self, value, fields, title=None, **kwargs):
-        self.value = value.strip()
+        self.value = value
         if self.value.startswith('-'):
             self.direction = self.DESC
         else:
             self.direction = self.ASC
-        if isinstance(fields, basestring):
+        if not isinstance(fields, (list, tuple)):
             self.fields = [fields]
         else:
             self.fields = fields
@@ -402,6 +538,7 @@ class OrderingValue(object):
     def apply(self, query, params):
         return query.order_by(*self.fields)
         
+
 class OrderingFilter(BaseFilter):
     def __init__(self, name, *args, **kwargs):
         self.name = name
@@ -421,9 +558,12 @@ class OrderingFilter(BaseFilter):
         
     def apply(self, query, params):
         ordering_value = None
-        orderlist = params.get(self.name)
-        if orderlist:
-            ordering_value = self.get_value(orderlist[0].strip())
+        valuelist = params.get(self.name, [])
+        if not isinstance(valuelist, (list, tuple)):
+            valuelist = [valuelist]
+
+        if valuelist:
+            ordering_value = self.get_value(valuelist[0])
 
         if not ordering_value:
             ordering_value = self.default_value
